// Copyright (C) 2010-2011 Institut de Robotica i Informatica Industrial, CSIC-UPC.
// Author 
// All rights reserved.
//
// This file is part of iri-ros-pkg
// iri-ros-pkg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 
// IMPORTANT NOTE: This code has been generated through a script from the 
// iri_ros_scripts. Please do NOT delete any comments to guarantee the correctness
// of the scripts. ROS topics can be easly add by using those scripts. Please
// refer to the IRI wiki page for more information:
// http://wikiri.upc.es/index.php/Robotics_Lab

#ifndef _iri_joints_client_bt_alg_node_h_
#define _iri_joints_client_bt_alg_node_h_

#include <iri_base_algorithm/iri_base_algorithm.h>
#include "iri_joints_client_bt_alg.h"

// [publisher subscriber headers]

// [service client headers]

// [nav server client headers]

#include <iri_joints_module/iri_joints_module_bt.h>
#include "iri_bt_factory.h"

#include "behaviortree_cpp_v3/loggers/bt_cout_logger.h"
#ifdef ZMQ_FOUND
#include "behaviortree_cpp_v3/loggers/bt_zmq_publisher.h"
#endif

/**
 * \brief IRI ROS Specific Algorithm Class
 *
 */
class IriJointsClientBtAlgNode : public algorithm_base::IriBaseAlgorithm<IriJointsClientBtAlgorithm>
{
  private:
    // [publisher attributes]

    // [subscriber attributes]

    // [service attributes]

    // [client attributes]

    // [nav server attributes]

    // [nav client attributes]

    CIriJointsModuleBT joints_module_bt;
    
    IriBehaviorTreeFactory factory;
    BT::Tree tree;
    BT::NodeStatus status;

    BT::StdCoutLogger   *logger_cout;
    #ifdef ZMQ_FOUND
    BT::PublisherZMQ *publisher_zmq;
    #endif

    std::string path;
    std::string tree_xml_file;

    bool send_ticks;

    /**
      * \brief Synchronized set move to angles parameters.
      *
      * This function sets join_names, angles and vel of each servo.
      *
      * It has the following output ports:
      *
      *   joint_names (std::vector<std::string>): Joint name of each servo.
      *
      *   angles (std::vecotr<double>): Each servo desired angle.
      *
      *   vel (std::vecotr<double>): Each servo desired velocity.
      *
      * \param self node with the required inputs defined as follows:
      *
      * \return a BT:NodeStatus indicating whether the request has been
      * successfull (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE).
      *
      */
    BT::NodeStatus set_move_angles_arguments(BT::TreeNode& self);

  public:
   /**
    * \brief Constructor
    * 
    * This constructor initializes specific class attributes and all ROS
    * communications variables to enable message exchange.
    */
    IriJointsClientBtAlgNode(void);

   /**
    * \brief Destructor
    * 
    * This destructor frees all necessary dynamic memory allocated within this
    * this class.
    */
    ~IriJointsClientBtAlgNode(void);

  protected:
   /**
    * \brief main node thread
    *
    * This is the main thread node function. Code written here will be executed
    * in every node loop while the algorithm is on running state. Loop frequency 
    * can be tuned by modifying loop_rate attribute.
    *
    * Here data related to the process loop or to ROS topics (mainly data structs
    * related to the MSG and SRV files) must be updated. ROS publisher objects 
    * must publish their data in this process. ROS client servers may also
    * request data to the corresponding server topics.
    */
    void mainNodeThread(void);

   /**
    * \brief dynamic reconfigure server callback
    * 
    * This method is called whenever a new configuration is received through
    * the dynamic reconfigure. The derivated generic algorithm class must 
    * implement it.
    *
    * \param config an object with new configuration from all algorithm 
    *               parameters defined in the config file.
    * \param level  integer referring the level in which the configuration
    *               has been changed.
    */
    void node_config_update(Config &config, uint32_t level);

   /**
    * \brief node add diagnostics
    *
    * In this abstract function additional ROS diagnostics applied to the 
    * specific algorithms may be added.
    */
    void addNodeDiagnostics(void);

    // [diagnostic functions]
    
    // [test functions]
};

#endif
