#include "iri_joints_client_bt_alg_node.h"

IriJointsClientBtAlgNode::IriJointsClientBtAlgNode(void) :
  algorithm_base::IriBaseAlgorithm<IriJointsClientBtAlgorithm>(),
  joints_module_bt()
{
  //init class attributes if necessary
  this->setRate(1);//in [Hz]

  // [init publishers]
  
  // [init subscribers]
  
  // [init services]
  
  // [init clients]
  
  std::string xml_path;
  if(!this->private_node_handle_.getParam("xml_path",xml_path))
    ROS_ERROR("IriJointsClientBtAlgNode::Constructor -> need to define xml_path parameter");
  else
    ROS_DEBUG_STREAM("IriJointsClientBtAlgNode::Constructor -> xml_path set to: " << xml_path);

  std::string xml_file;
  if(!this->private_node_handle_.getParam("xml_file",xml_file))
    ROS_ERROR("IriJointsClientBtAlgNode::Constructor -> need to define xml_file parameter");
  else
    ROS_DEBUG_STREAM("IriJointsClientBtAlgNode::Constructor -> xml_file set to: " << xml_file);

  this->joints_module_bt.init(this->factory);
  BT::PortsList set_move_param_ports = {BT::OutputPort<std::vector<std::string>>("joint_names"), BT::OutputPort<std::vector<double>>("angles"), BT::OutputPort<std::vector<double>>("vel")};
  this->factory.registerSimpleAction("set_move_angles_arguments",  std::bind(&IriJointsClientBtAlgNode::set_move_angles_arguments, this, std::placeholders::_1), set_move_param_ports);
  
  std::cout << xml_path << "/" << xml_file << std::endl;
  try
  {
    this->tree = this->factory.createTreeFromFile(xml_path + "/" + xml_file);
  }
  catch(BT::RuntimeError &e)
  {
    ROS_ERROR_STREAM("IriJointsClientBtAlgNode::Constructor -> Exception caught: " << e.what());
  }

  this->logger_cout      = new BT::StdCoutLogger(this->tree);
  // this->logger_minitrace = new BT::MinitraceLogger(this->tree, (this->path + "/logs/" + this->tree_xml_file + ".json").c_str());
  // this->logger_file      = new BT::FileLogger(this->tree, (this->path + "/logs/" + this->tree_xml_file + ".fbl").c_str());
  #ifdef ZMQ_FOUND
  this->publisher_zmq    = new BT::PublisherZMQ(this->tree);
  #endif
  // BT::printTreeRecursively(this->tree.root_node);  
}

IriJointsClientBtAlgNode::~IriJointsClientBtAlgNode(void)
{
  // [free dynamic memory]
  if (this->logger_cout != NULL)
  {
    delete this->logger_cout;
    this->logger_cout = NULL;
  }
  // delete this->logger_minitrace;
  // this->logger_minitrace = NULL;
  // delete this->logger_file;
  // this->logger_file = NULL;
  #ifdef ZMQ_FOUND
  if (this->publisher_zmq != NULL)
  {
    delete this->publisher_zmq;
    this->publisher_zmq = NULL;
  }
  #endif
}

void IriJointsClientBtAlgNode::mainNodeThread(void)
{
  // [fill msg structures]
  
  // [fill srv structure and make request to the server]

  // [publish messages]
  static bool running = true;
  try
  {
    if (send_ticks && running)
    {
      ROS_INFO("---Tick---");
      this->status = this->tree.tickRoot();
      if (this->status != BT::NodeStatus::RUNNING)
        running = false;
      ROS_INFO_STREAM("Status: " << this->status);
    }
  }
  catch(BT::LogicError &e)
  {
    ROS_ERROR_STREAM("IriJointsClientBtAlgNode::mainNodeThread -> EXCEPTION caught:" << e.what());
  }
}

BT::NodeStatus IriJointsClientBtAlgNode::set_move_angles_arguments(BT::TreeNode& self)
{
  std::vector<std::string> joint_names;
  std::vector<double> angles, vel;

  joint_names.push_back("pan");
  joint_names.push_back("tilt");
  angles.push_back(0.0);
  angles.push_back(0.0);
  vel.push_back(0.5);
  vel.push_back(0.5);

  self.setOutput("joint_names", joint_names);
  self.setOutput("angles", angles);
  self.setOutput("vel", vel);
  return BT::NodeStatus::SUCCESS;
}

/*  [subscriber callbacks] */

/*  [service callbacks] */

/*  [nav callbacks] */

void IriJointsClientBtAlgNode::node_config_update(Config &config, uint32_t level)
{
  this->alg_.lock();
  send_ticks = config.send_ticks;
  this->alg_.unlock();
}

void IriJointsClientBtAlgNode::addNodeDiagnostics(void)
{
}

/* main function */
int main(int argc,char *argv[])
{
  return algorithm_base::main<IriJointsClientBtAlgNode>(argc, argv, "iri_joints_client_bt_alg_node");
}
