#include <iri_joints_module/iri_joints_module_bt.h>
#include <iri_joints_module/iri_joints_module.h>
#include <stdlib.h> 

CIriJointsModuleBT::CIriJointsModuleBT() :
  joints_module("joints_module",ros::this_node::getName())
{

}

CIriJointsModuleBT::CIriJointsModuleBT(std::string &module_name) :
  joints_module(module_name,ros::this_node::getName())
{

}

void CIriJointsModuleBT::init(IriBehaviorTreeFactory &factory)
{
	//Definition of ports
  BT::PortsList point_to_ports = {BT::InputPort<double>("x"), BT::InputPort<double>("y"), BT::InputPort<double>("z"), BT::InputPort<double>("max_vel")};
  BT::PortsList async_point_to_ports = {BT::InputPort<double>("x"), BT::InputPort<double>("y"), BT::InputPort<double>("z"), BT::InputPort<double>("max_vel"), BT::BidirectionalPort<bool>("new_goal")};

  BT::PortsList move_to_angles_ports = {BT::InputPort<std::vector<std::string>>("joint_names"), BT::InputPort<std::vector<double>>("angles"), BT::InputPort<std::vector<double>>("vel")};
  BT::PortsList async_move_to_angles_ports = {BT::InputPort<std::vector<std::string>>("joint_names"), BT::InputPort<std::vector<double>>("angles"), BT::InputPort<std::vector<double>>("vel"), BT::BidirectionalPort<bool>("new_goal")};
  
  BT::PortsList update_target_ports = {BT::InputPort<double>("x"), BT::InputPort<double>("y"), BT::InputPort<double>("z")};
  
  BT::PortsList set_tolerance_ports = {BT::InputPort<double>("tol")};
  BT::PortsList set_base_frame_ports = {BT::InputPort<std::string>("base_frame")};
  BT::PortsList set_urdf_param_ports = {BT::InputPort<std::string>("urdf_param")};

  BT::PortsList target_update_point_port = {BT::InputPort<geometry_msgs::PointStamped>("point"), BT::BidirectionalPort<bool>("new_goal")};

  //Registration of conditions
  factory.registerSimpleCondition("is_point_to_finished",  std::bind(&CIriJointsModuleBT::is_point_to_finished, this));
  factory.registerSimpleCondition("is_move_to_angles_finished",  std::bind(&CIriJointsModuleBT::is_move_to_angles_finished, this));
  factory.registerSimpleCondition("is_point_to_tracker_finished",  std::bind(&CIriJointsModuleBT::is_tracker_finished, this));

  factory.registerSimpleCondition("is_joints_module_running",  std::bind(&CIriJointsModuleBT::is_joints_module_running, this));
  factory.registerSimpleCondition("is_joints_module_success",  std::bind(&CIriJointsModuleBT::is_joints_module_success, this));
  factory.registerSimpleCondition("is_joints_module_action_start_fail",  std::bind(&CIriJointsModuleBT::is_joints_module_action_start_fail, this));
  factory.registerSimpleCondition("is_joints_module_timeout",  std::bind(&CIriJointsModuleBT::is_joints_module_timeout, this));
  factory.registerSimpleCondition("is_joints_module_fb_watchdog",  std::bind(&CIriJointsModuleBT::is_joints_module_fb_watchdog, this));
  factory.registerSimpleCondition("is_joints_module_aborted",  std::bind(&CIriJointsModuleBT::is_joints_module_aborted, this));
  factory.registerSimpleCondition("is_joints_module_preempted",  std::bind(&CIriJointsModuleBT::is_joints_module_preempted, this));
  factory.registerSimpleCondition("is_joints_module_rejected",  std::bind(&CIriJointsModuleBT::is_joints_module_rejected, this));
  factory.registerSimpleCondition("is_joints_module_set_param_fail",  std::bind(&CIriJointsModuleBT::is_joints_module_set_param_fail, this));
  factory.registerSimpleCondition("is_joints_module_param_not_present",  std::bind(&CIriJointsModuleBT::is_joints_module_param_not_present, this));
  factory.registerSimpleCondition("is_joints_module_no_joint_states",  std::bind(&CIriJointsModuleBT::is_joints_module_no_joint_states, this));
  factory.registerSimpleCondition("is_joints_module_target_lost",  std::bind(&CIriJointsModuleBT::is_joints_module_target_lost, this));

  //Registration of actions
  factory.registerSimpleAction("sync_point_to",  std::bind(&CIriJointsModuleBT::sync_point_to, this, std::placeholders::_1), point_to_ports);
  factory.registerIriAsyncAction("async_point_to",  std::bind(&CIriJointsModuleBT::async_point_to, this, std::placeholders::_1), async_point_to_ports);
  factory.registerSimpleAction("cancel_point_to",  std::bind(&CIriJointsModuleBT::cancel_point_to, this));
  factory.registerIriAsyncAction("async_is_point_to_finished",  std::bind(&CIriJointsModuleBT::async_is_point_to_finished, this));

  factory.registerSimpleAction("sync_move_to_angles",  std::bind(&CIriJointsModuleBT::sync_move_to_angles, this, std::placeholders::_1), move_to_angles_ports);
  factory.registerIriAsyncAction("async_move_to_angles",  std::bind(&CIriJointsModuleBT::async_move_to_angles, this, std::placeholders::_1), async_move_to_angles_ports);
  factory.registerSimpleAction("cancel_move_to_angles",  std::bind(&CIriJointsModuleBT::cancel_move_to_angles, this));
  factory.registerIriAsyncAction("async_is_move_to_angles_finished",  std::bind(&CIriJointsModuleBT::async_is_move_to_angles_finished, this));

  factory.registerSimpleAction("point_to_tracker_update_target",  std::bind(&CIriJointsModuleBT::target_update, this, std::placeholders::_1), update_target_ports);
  factory.registerSimpleAction("enable_point_to_tracker",  std::bind(&CIriJointsModuleBT::enable_tracker, this));
  factory.registerSimpleAction("cancel_point_to_tracker",  std::bind(&CIriJointsModuleBT::cancel_tracker, this));
  factory.registerIriAsyncAction("async_is_point_to_tracker_finished",  std::bind(&CIriJointsModuleBT::async_is_tracker_finished, this));

  factory.registerSimpleAction("set_point_to_inv_kin_solver_tolerance",  std::bind(&CIriJointsModuleBT::set_tolerance, this, std::placeholders::_1), set_tolerance_ports);
  factory.registerSimpleAction("set_point_to_kin_chain_base_frame",  std::bind(&CIriJointsModuleBT::set_base_frame, this, std::placeholders::_1), set_base_frame_ports);
  factory.registerSimpleAction("set_point_to_kin_chain_urdf_param",  std::bind(&CIriJointsModuleBT::set_urdf_param, this, std::placeholders::_1), set_urdf_param_ports);
  factory.registerSimpleAction("point_to_tracker_target_update_pointstamped",  std::bind(&CIriJointsModuleBT::target_update_pointstamped, this, std::placeholders::_1), target_update_point_port);
}

CIriJointsModuleBT::~CIriJointsModuleBT(void)
{
  // [free dynamic memory]
}

////////////////////////////////////////////////////////////// Actions //////////////////////////////////////////////////////////////
BT::NodeStatus CIriJointsModuleBT::sync_point_to(BT::TreeNode& self)
{
  ROS_DEBUG("CIriJointsModuleBT::sync_point_to-> sync_point_to");
  BT::Optional<double> x = self.getInput<double>("x");
  BT::Optional<double> y = self.getInput<double>("y");
  BT::Optional<double> z = self.getInput<double>("z");
  BT::Optional<double> max_vel = self.getInput<double>("max_vel");

  if (!x || !y || !z || !max_vel)
  {
    ROS_ERROR("CIriJointsModuleBT::sync_point_to-> Incorrect or missing input. It needs the following input ports: x(double), y(double), z(double) and max_vel(double)");
    return BT::NodeStatus::FAILURE;
  }

  double x_goal, y_goal, z_goal, max_vel_goal;
  x_goal = x.value();
  y_goal = y.value();
  z_goal = z.value();
  max_vel_goal = max_vel.value();

  this->joints_module.point_to(x_goal, y_goal, z_goal, max_vel_goal);
  return BT::NodeStatus::SUCCESS;
}

BT::NodeStatus CIriJointsModuleBT::async_point_to(BT::TreeNode& self)
{
  ROS_DEBUG("CIriJointsModuleBT::async_point_to-> async_point_to");
  BT::Optional<double> x = self.getInput<double>("x");
  BT::Optional<double> y = self.getInput<double>("y");
  BT::Optional<double> z = self.getInput<double>("z");
  BT::Optional<double> max_vel = self.getInput<double>("max_vel");
  BT::Optional<bool> new_goal_in = self.getInput<bool>("new_goal");

  bool new_goal;
  if (!new_goal_in)
    new_goal = false;
  else
    new_goal = new_goal_in.value();

  if (new_goal)
  {
    if (!x || !y || !z || !max_vel)
    {
      ROS_ERROR("CIriJointsModuleBT::async_point_to-> Incorrect or missing input. It needs the following input ports: x(double), y(double), z(double) and max_vel(double)");
      return BT::NodeStatus::FAILURE;
    }

    double x_goal, y_goal, z_goal, max_vel_goal;
    x_goal = x.value();
    y_goal = y.value();
    z_goal = z.value();
    max_vel_goal = max_vel.value();

    self.setOutput("new_goal", false);
    this->joints_module.point_to(x_goal, y_goal, z_goal, max_vel_goal);
  }
  if (this->joints_module.is_point_to_finished())
  {
    iri_joints_module_status_t joints_module_status = this->joints_module.get_status();
    if (joints_module_status == IRI_JOINTS_MODULE_SUCCESS)
      return BT::NodeStatus::SUCCESS;
    else
      return BT::NodeStatus::FAILURE;
  }
  return BT::NodeStatus::RUNNING;
}

BT::NodeStatus CIriJointsModuleBT::cancel_point_to(void)
{
  ROS_DEBUG("CIriJointsModuleBT::cancel_point_to-> cancel_point_to");
  this->joints_module.cancel_point_to();
  return BT::NodeStatus::SUCCESS;
}

BT::NodeStatus CIriJointsModuleBT::async_is_point_to_finished(void)
{
  if (this->joints_module.is_point_to_finished())
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::RUNNING;
}


BT::NodeStatus CIriJointsModuleBT::sync_move_to_angles(BT::TreeNode& self)
{
  ROS_DEBUG("CIriJointsModuleBT::sync_move_to_angles-> sync_move_to_angles");
  BT::Optional<std::vector<std::string>> joint_names = self.getInput<std::vector<std::string>>("joint_names");
  BT::Optional<std::vector<double>> angles = self.getInput<std::vector<double>>("angles");
  BT::Optional<std::vector<double>> vel = self.getInput<std::vector<double>>("vel");

  if (!joint_names || !angles || !vel)
  {
    ROS_ERROR("CIriJointsModuleBT::sync_move_to_angles-> Incorrect or missing input. It needs the following input ports: joint_names(std::vector<std::String>), angle(std::vector<double>) and vel(std::vector<double>).");
    return BT::NodeStatus::FAILURE;
  }

  std::vector<std::string> joint_names_vect;
  std::vector<double> angles_vect, vel_vect;
  joint_names_vect = joint_names.value();
  angles_vect = angles.value();
  vel_vect = vel.value();

  this->joints_module.move_to_angles(joint_names_vect, angles_vect, vel_vect);
  return BT::NodeStatus::SUCCESS;
}

BT::NodeStatus CIriJointsModuleBT::async_move_to_angles(BT::TreeNode& self)
{
  ROS_DEBUG("CIriJointsModuleBT::async_move_to_angles-> async_move_to_angles");
  BT::Optional<std::vector<std::string>> joint_names = self.getInput<std::vector<std::string>>("joint_names");
  BT::Optional<std::vector<double>> angles = self.getInput<std::vector<double>>("angles");
  BT::Optional<std::vector<double>> vel = self.getInput<std::vector<double>>("vel");
  BT::Optional<bool> new_goal_in = self.getInput<bool>("new_goal");

  bool new_goal;
  if (!new_goal_in)
    new_goal = false;
  else
    new_goal = new_goal_in.value();

  if (new_goal)
  {
    if (!joint_names || !angles || !vel)
    {
      ROS_ERROR("CIriJointsModuleBT::async_move_to_angles-> Incorrect or missing input. It needs the following input ports: joint_names(std::vector<std::String>), angle(std::vector<double>) and vel(std::vector<double>).");
      return BT::NodeStatus::FAILURE;
    }

    std::vector<std::string> joint_names_vect;
    std::vector<double> angles_vect, vel_vect;
    joint_names_vect = joint_names.value();
    angles_vect = angles.value();
    vel_vect = vel.value();

    self.setOutput("new_goal", false);
    this->joints_module.move_to_angles(joint_names_vect, angles_vect, vel_vect);
  }
  if (this->joints_module.is_move_to_angles_finished())
  {
    iri_joints_module_status_t joints_module_status = this->joints_module.get_status();
    if (joints_module_status == IRI_JOINTS_MODULE_SUCCESS)
      return BT::NodeStatus::SUCCESS;
    else
      return BT::NodeStatus::FAILURE;
  }
  return BT::NodeStatus::RUNNING;
}

BT::NodeStatus CIriJointsModuleBT::cancel_move_to_angles(void)
{
  ROS_DEBUG("CIriJointsModuleBT::cancel_move_to_angles-> cancel_move_to_angles");
  this->joints_module.cancel_move_to_angles();
  return BT::NodeStatus::SUCCESS;
}

BT::NodeStatus CIriJointsModuleBT::async_is_move_to_angles_finished(void)
{
  if (this->joints_module.is_move_to_angles_finished())
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::RUNNING;
}

BT::NodeStatus CIriJointsModuleBT::target_update(BT::TreeNode& self)
{
  ROS_DEBUG("CIriJointsModuleBT::target_update-> target_update");
  BT::Optional<double> x = self.getInput<double>("x");
  BT::Optional<double> y = self.getInput<double>("y");
  BT::Optional<double> z = self.getInput<double>("z");

  if (!x || !y || !z)
  {
    ROS_ERROR("CIriJointsModuleBT::target_update-> Incorrect or missing input. It needs the following input ports: x(double), y(double) and z(double)");
    return BT::NodeStatus::FAILURE;
  }

  double x_goal, y_goal, z_goal;
  x_goal = x.value();
  y_goal = y.value();
  z_goal = z.value();

  this->joints_module.target_update(x_goal, y_goal, z_goal);
  return BT::NodeStatus::SUCCESS;
}

BT::NodeStatus CIriJointsModuleBT::target_update_pointstamped(BT::TreeNode& self)
{
  ROS_DEBUG("CIriJointsModuleBT::target_update_pointstamped-> target_update");
  BT::Optional<geometry_msgs::PointStamped> point = self.getInput<geometry_msgs::PointStamped>("point");
  BT::Optional<bool> new_goal_in = self.getInput<bool>("new_goal");

  bool new_goal;
  if (!new_goal_in)
    new_goal = false;
  else
    new_goal = new_goal_in.value();

  if (new_goal)
  {
    if (!point)
    {
      ROS_ERROR("CIriJointsModuleBT::target_update_pointstamped-> Incorrect or missing input. It needs the following input ports: point(geometry_msgs::PointStamped)");
      return BT::NodeStatus::FAILURE;
    }

    geometry_msgs::PointStamped point_value = point.value();

    this->joints_module.target_update(point_value);
  }
  if (new_goal)
    self.setOutput("new_goal", false);
  return BT::NodeStatus::SUCCESS;
}

BT::NodeStatus CIriJointsModuleBT::enable_tracker(void)
{
  ROS_DEBUG("CIriJointsModuleBT::enable_tracker-> enable_tracker");
  if (this->joints_module.enable_tracker())
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE; 
}

BT::NodeStatus CIriJointsModuleBT::cancel_tracker(void)
{
  ROS_DEBUG("CIriJointsModuleBT::cancel_tracker-> cancel_tracker");
  this->joints_module.cancel_tracker();
  return BT::NodeStatus::SUCCESS;
}

BT::NodeStatus CIriJointsModuleBT::async_is_tracker_finished(void)
{
  if (this->joints_module.is_tracker_finished())
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::RUNNING;
}


BT::NodeStatus CIriJointsModuleBT::set_tolerance(BT::TreeNode& self)
{
  ROS_DEBUG("CIriJointsModuleBT::set_tolerance-> set_tolerance");
  BT::Optional<double> tolerance = self.getInput<double>("tol");

  if (!tolerance)
  {
    ROS_ERROR("CIriJointsModuleBT::set_tolerance-> Incorrect or missing input. It needs the following input ports: tol(double)");
    return BT::NodeStatus::FAILURE;
  }

  double tol;
  tol = tolerance.value();
  if (this->joints_module.set_tolerance(tol))
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

BT::NodeStatus CIriJointsModuleBT::set_base_frame(BT::TreeNode& self)
{
  ROS_DEBUG("CIriJointsModuleBT::set_base_frame-> set_base_frame");
  BT::Optional<std::string> base_frame_in = self.getInput<std::string>("base_frame");

  if (!base_frame_in)
  {
    ROS_ERROR("CIriJointsModuleBT::set_base_frame-> Incorrect or missing input. It needs the following input ports: base_frame(std::string)");
    return BT::NodeStatus::FAILURE;
  }

  std::string base_frame;
  base_frame = base_frame_in.value();
  if (this->joints_module.set_base_frame(base_frame))
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

BT::NodeStatus CIriJointsModuleBT::set_urdf_param(BT::TreeNode& self)
{
  ROS_DEBUG("CIriJointsModuleBT::set_urdf_param-> set_urdf_param");
  BT::Optional<std::string> urdf_param_in = self.getInput<std::string>("urdf_param");

  if (!urdf_param_in)
  {
    ROS_ERROR("CIriJointsModuleBT::set_urdf_param-> Incorrect or missing input. It needs the following input ports: urdf_param(std::string)");
    return BT::NodeStatus::FAILURE;
  }

  std::string urdf_param;
  urdf_param = urdf_param_in.value();
  if (this->joints_module.set_urdf_param(urdf_param))
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

////////////////////////////////////////////////////////////// Conditions //////////////////////////////////////////////////////////////

BT::NodeStatus CIriJointsModuleBT::is_point_to_finished(void)
{
  if (this->joints_module.is_point_to_finished())
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

BT::NodeStatus CIriJointsModuleBT::is_move_to_angles_finished(void)
{
  if (this->joints_module.is_move_to_angles_finished())
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

BT::NodeStatus CIriJointsModuleBT::is_tracker_finished(void)
{
  if (this->joints_module.is_tracker_finished())
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}


BT::NodeStatus CIriJointsModuleBT::is_joints_module_running(void)
{
  iri_joints_module_status_t joints_module_status = this->joints_module.get_status();

  if (joints_module_status == IRI_JOINTS_MODULE_RUNNING)
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

BT::NodeStatus CIriJointsModuleBT::is_joints_module_success(void)
{
  iri_joints_module_status_t joints_module_status = this->joints_module.get_status();

  if (joints_module_status == IRI_JOINTS_MODULE_SUCCESS)
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

BT::NodeStatus CIriJointsModuleBT::is_joints_module_action_start_fail(void)
{
  iri_joints_module_status_t joints_module_status = this->joints_module.get_status();

  if (joints_module_status == IRI_JOINTS_MODULE_ACTION_START_FAIL)
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

BT::NodeStatus CIriJointsModuleBT::is_joints_module_timeout(void)
{
  iri_joints_module_status_t joints_module_status = this->joints_module.get_status();

  if (joints_module_status == IRI_JOINTS_MODULE_TIMEOUT)
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

BT::NodeStatus CIriJointsModuleBT::is_joints_module_fb_watchdog(void)
{
  iri_joints_module_status_t joints_module_status = this->joints_module.get_status();

  if (joints_module_status == IRI_JOINTS_MODULE_FB_WATCHDOG)
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

BT::NodeStatus CIriJointsModuleBT::is_joints_module_aborted(void)
{
  iri_joints_module_status_t joints_module_status = this->joints_module.get_status();

  if (joints_module_status == IRI_JOINTS_MODULE_ABORTED)
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}


BT::NodeStatus CIriJointsModuleBT::is_joints_module_preempted(void)
{
  iri_joints_module_status_t joints_module_status = this->joints_module.get_status();

  if (joints_module_status == IRI_JOINTS_MODULE_PREEMPTED)
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

BT::NodeStatus CIriJointsModuleBT::is_joints_module_rejected(void)
{
  iri_joints_module_status_t joints_module_status = this->joints_module.get_status();

  if (joints_module_status == IRI_JOINTS_MODULE_REJECTED)
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

BT::NodeStatus CIriJointsModuleBT::is_joints_module_set_param_fail(void)
{
  iri_joints_module_status_t joints_module_status = this->joints_module.get_status();

  if (joints_module_status == IRI_JOINTS_MODULE_SET_PARAM_FAIL)
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

BT::NodeStatus CIriJointsModuleBT::is_joints_module_param_not_present(void)
{
  iri_joints_module_status_t joints_module_status = this->joints_module.get_status();

  if (joints_module_status == IRI_JOINTS_MODULE_PARAM_NOT_PRESENT)
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

BT::NodeStatus CIriJointsModuleBT::is_joints_module_no_joint_states(void)
{
  iri_joints_module_status_t joints_module_status = this->joints_module.get_status();

  if (joints_module_status == IRI_JOINTS_MODULE_NO_JOINT_STATES)
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}

BT::NodeStatus CIriJointsModuleBT::is_joints_module_target_lost(void)
{
  iri_joints_module_status_t joints_module_status = this->joints_module.get_status();

  if (joints_module_status == IRI_JOINTS_MODULE_TARGET_LOST)
    return BT::NodeStatus::SUCCESS;
  return BT::NodeStatus::FAILURE;
}